/*
 * ftp.h
 */

/*
 * FTP global variables.
 */

/*
 * Options and other state info.
 */
int	trace;			/* trace packets exchanged */
int	hash;			/* print # for each buffer transferred */
int	sendport;		/* use PORT cmd for each data connection */
int	verbose;		/* print messages coming back from server */
int	connected;		/* connected to server */
int	fromatty;		/* input is from a terminal */
int	interactive;		/* interactively prompt on m* cmds */
int	debug;			/* debugging level */
int	bell;			/* ring bell on cmd completion */
int	doglob;			/* glob local file names */
int	autologin;		/* establish user account on connection */
int	proxy;			/* proxy server connection active */
int	proxflag;		/* proxy connection exists */
int	sunique;		/* store files on server with unique name */
int	runique;		/* store local files with unique name */
int	mcase;			/* map upper to lower case for mget names */
int	ntflag;			/* use ntin ntout tables for name translation */
int	mapflag;		/* use mapin mapout templates on file names */
int	code;			/* return/reply code for ftp command */
int	crflag;			/* if 1, strip car. rets. on ascii gets */
char	pasv[64];		/* passive port for proxy data connection */
char	*altarg;		/* argv[1] with no shell-like preprocessing  */
char	ntin[17];		/* input translation table */
char	ntout[17];		/* output translation table */
#include <sys/param.h>
char	mapin[MAXPATHLEN];	/* input map template */
char	mapout[MAXPATHLEN];	/* output map template */
char	typename[32];		/* name of file transfer type */
int	type;			/* requested file transfer type */
int	curtype;		/* current file transfer type */
char	structname[32];		/* name of file transfer structure */
int	stru;			/* file transfer structure */
char	formname[32];		/* name of file transfer format */
int	form;			/* file transfer format */
char	modename[32];		/* name of file transfer mode */
int	mode;			/* file transfer mode */
char	bytename[32];		/* local byte size in ascii */
int	bytesize;		/* local byte size in binary */

char	*hostname;		/* name of host connected to */
int	unix_server;		/* server is unix, can use binary for ascii */
int	unix_proxy;		/* proxy is unix, can use binary for ascii */

struct	servent *sp;		/* service spec for tcp/ftp */

#include <setjmp.h>
jmp_buf	toplevel;		/* non-local goto stuff for cmd scanner */

char	line[200];		/* input line buffer */
char	*stringbase;		/* current scan point in line buffer */
char	argbuf[200];		/* argument storage buffer */
char	*argbase;		/* current storage point in arg buffer */
int	margc;			/* count of arguments on input line */
char	*margv[20];		/* args parsed from input line */
int     cpend;                  /* flag: if != 0, then pending server reply */
int	mflag;			/* flag: if != 0, then active multi command */

int	options;		/* used during socket creation */

/*
 * Format of command table.
 */
struct cmd {
	char	*c_name;	/* name of command */
	char	*c_help;	/* help string */
	char	c_bell;		/* give bell when command completes */
	char	c_conn;		/* must be connected to use command */
	char	c_proxy;	/* proxy server may execute */
	int	(*c_handler)();	/* function to call */
};

struct macel {
	char mac_name[9];	/* macro name */
	char *mac_start;	/* start of macro in macbuf */
	char *mac_end;		/* end of macro in macbuf */
};

int macnum;			/* number of defined macros */
struct macel macros[16];
char macbuf[4096];


extern char *globerr;
extern char *home;
extern off_t restart_point;
extern char reply_string[];


/*
 * public functions
 */

/* getsstr.c */

public char *getsstr PROTO((char *));


/* glob.c */

public char **ftpglob PROTO((char *));
public void blkfree PROTO((char **));


/* cmds.c */

public int another PROTO((int *, char ***, char *));
public int setpeer PROTO((int, char **));
public int settype PROTO((int, char **));
public void changetype PROTO((int, int));
public int setbinary PROTO((void));
public int setascii PROTO((void));
public int settenex PROTO((void));
public int setmode2 PROTO((int, char **));
public int setform PROTO((int, char **));
public int setstruct PROTO((int, char **));
public int put PROTO((int, char **));
public int mput PROTO((int, char **));
public int reget PROTO((int, char **));
public int get PROTO((int, char **));
public int getit PROTO((int, char **, int, char *));
public int mget PROTO((int, char **));
public int status PROTO((int, char **));
public int setbell PROTO((void));
public int settrace PROTO((void));
public int sethash PROTO((void));
public int setverbose PROTO((void));
public int setport PROTO((void));
public int setprompt PROTO((void));
public int setglob PROTO((void));
public int setdebug PROTO((int, char **));
public int cd PROTO((int, char **));
public int lcd PROTO((int, char **));
public int delete PROTO((int, char **));
public int mdelete PROTO((int, char **));
public int renamefile PROTO((int, char **));
public int ls PROTO((int, char **));
public int  mls PROTO((int, char **));
public int shell PROTO((int, char **));
public int user PROTO((int, char **));
public int pwd PROTO((void));
public int makedir PROTO((int, char **));
public int removedir PROTO((int, char **));
public int quote PROTO((int, char **));
public int site PROTO((int, char **));
public void quote1 PROTO((char *, int, char **));
public int do_chmod PROTO((int, char **));
public int do_umask PROTO((int, char **));
public int doidle PROTO((int, char **));
public int rmthelp PROTO((int, char **));
public int quit PROTO((void));
public int disconnect PROTO((void));
public int confirm PROTO((char *, char *));
public void fatal PROTO((char *));
public int globulize PROTO((char **));
public int account PROTO((int, char **));
public int doproxy PROTO((int, char **));
public int setcase PROTO((void));
public int setcr PROTO((void));
public int setntrans PROTO((int, char **));
public int setnmap PROTO((int, char **));
public int setsunique PROTO((void));
public int setrunique PROTO((void));
public int cdup PROTO((void));
public int restart PROTO((int, char **));
public int syst PROTO((void));
public int macdef PROTO((int, char **));
public int sizecmd PROTO((int, char **));
public int modtime PROTO((int, char **));
public int rmtstatus PROTO((int, char **));
public int newer PROTO((int, char **));


/* domacro.c */

public int domacro PROTO((int, char **));


/* main.c */

public RETSIGTYPE intr(); /* signal */
public RETSIGTYPE lostpeer (); /* signal */
public struct cmd *getcmd PROTO((char *));
public void makeargv PROTO((void));
public int help PROTO((int, char **));


/* ftp.c */

public char *hookup PROTO((char *, int));
public int login PROTO((char *));
public int command (); /* xxx */
public int getreply PROTO((int));
public void sendrequest PROTO((char *, char *, char *, int));
public void recvrequest PROTO((char *, char *, char *, char *, int));
public void pswitch PROTO((int));
public int reset PROTO((void));

/* ruserpass.c */

public int ruserpass PROTO((char *, char **, char **, char **));


/*
 * Copyright (c) 1985, 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)ftp_var.h	5.9 (Berkeley) 6/1/90
 */
